import itertools
import string
from dataclasses import dataclass
from typing import Sequence, SupportsIndex

from rich.color import Color
from typing_extensions import Self

from .colormap import Tag


@dataclass
class RGB:
    """
    Dataclass representing a 3 channel RGB color that is converted to a `rich` tag when casted to a string.

    >>> color = RGB(100, 100, 100)
    >>> str(color)
    >>> "[rgb(100,100,100)]"
    >>> from rich.console import Console
    >>> console = Console()
    >>> console.print(f"{color}Yeehaw")

    Can also be initialized using a color name from https://rich.readthedocs.io/en/stable/appendix/colors.html

    >>> color = RGB(name="magenta3")
    >>> print(color)
    >>> "[rgb(215,0,215)]"

    Supports addition and subtraction of `RGB` objects as well as scalar multiplication and division.

    >>> color1 = RGB(100, 100, 100)
    >>> color2 = RGB(25, 50, 75)
    >>> print(color1 + color2)
    >>> "[rgb(125,150,175)]"
    >>> print(color2 * 2)
    >>> "[rgb(50,100,150)]"
    """

    # Typing these as floats so `Gradient` can fractionally increment them
    # When casted to a string, the values will be rounded to integers
    r: float = 0
    g: float = 0
    b: float = 0
    name: str = ""

    def __post_init__(self):
        if self.name:
            self.r, self.g, self.b = Color.parse(self.name).get_truecolor()

    def __str__(self) -> str:
        return f"[rgb({round(self.r)},{round(self.g)},{round(self.b)})]"

    def __sub__(self, other: Self) -> Self:
        return self.__class__(self.r - other.r, self.g - other.g, self.b - other.b)

    def __add__(self, other: Self) -> Self:
        return self.__class__(self.r + other.r, self.g + other.g, self.b + other.b)

    def __truediv__(self, val: float) -> Self:
        return self.__class__(self.r / val, self.g / val, self.b / val)

    def __mul__(self, val: float) -> Self:
        return self.__class__(self.r * val, self.g * val, self.b * val)

    def __eq__(self, other: object) -> bool:
        return all(getattr(self, c) == getattr(other, c) for c in "rgb")


ColorType = RGB | tuple[int, int, int] | str | Tag


class _Blender:
    """
    Apply a color blend from a start color to a stop color across text when printed with the `rich` package.
    """

    def __init__(
        self,
        start: RGB,
        stop: RGB,
    ):
        self.start = start
        self.stop = stop

    @property
    def valid_characters(self) -> str:
        """Characters a color step can be applied to."""
        return string.ascii_letters + string.digits + string.punctuation

    def _get_step_sizes(self, num_steps: int) -> RGB:
        """Returns a `RGB` object representing the step size for each color channel."""
        return (self.stop - self.start) / num_steps

    def _get_blended_color(self, step: int, step_sizes: RGB) -> RGB:
        """Returns a `RGB` object representing the color at `step`."""
        return self.start + (step_sizes * step)

    def _get_num_steps(self, text: str) -> int:
        """Returns the number of steps the blend should be divided into."""
        return len([ch for ch in text if ch in self.valid_characters]) - 1

    def apply(self, text: str) -> str:
        """Apply the blend to ascii letters, digits, and punctuation in `text`."""
        num_steps = self._get_num_steps(text)
        if num_steps < 0:  # no valid characters
            return text
        elif num_steps == 0:  # one valid character, just apply start color
            return f"{self.start}{text}[/]"
        step_sizes = self._get_step_sizes(num_steps)
        blended_text = ""
        step = 0
        for ch in text:
            if ch in self.valid_characters:
                blended_text += f"{self._get_blended_color(step, step_sizes)}{ch}[/]"
                step += 1
            else:
                blended_text += ch
        return blended_text


class Gradient(list[RGB]):
    """
    Apply an arbitrary number of color gradients to strings when using `rich`.

    When applied to a string, each character will increment in color from a start to a stop color.

    Colors can be specified by either
    a 3 tuple representing RGB values,
    a `pocketchange.RGB` object,
    a `pocketchange.Tag` object,
    or a color name from https://rich.readthedocs.io/en/stable/appendix/colors.html.

    Tuple:
    >>> gradient = Gradient([(255, 0, 0), (0, 255, 0)])

    `pocketchange.RGB`:
    >>> gradient = Gradient([RGB(255, 0, 0), RGB(0, 255, 0)])

    `pocketchange.Tag`:
    >>> colors = pocketchange.ColorMap()
    >>> gradient = Gradient([colors.red, colors.green])

    Name:
    >>> gradient = Gradient(["red", "green"])

    Usage:
    >>> from pocketchange import Gradient
    >>> from rich.console import Console
    >>>
    >>> console = Console()
    >>> gradient = Gradient(["red", "green"])
    >>> text = "Yeehaw"
    >>> gradient_text = gradient.apply(text)
    >>> # This produces:
    >>> print(gradient_text)
    >>> "[rgb(128,0,0)]Y[/][rgb(102,25,0)]e[/][rgb(76,51,0)]e[/][rgb(51,76,0)]h[/][rgb(25,102,0)]a[/][rgb(0,128,0)]w[/]"
    >>>
    >>> # When used with `console.print`, each character will be a different color
    >>> console.print(gradient_text)
    >>>
    >>> # `Gradient` inherits from `list` so colors may be appended, inserted, or extended
    >>> gradient.append("blue")
    >>> print(gradient.apply(text))
    >>> "[rgb(128,0,0)]Y[/][rgb(64,64,0)]e[/][rgb(0,128,0)]e[/][rgb(0,128,0)]h[/][rgb(0,64,64)]a[/][rgb(0,0,128)]w[/]"
    >>> print(gradient)
    >>> [RGB(r=128, g=0, b=0, name='red'), RGB(r=0, g=128, b=0, name='green'), RGB(r=0, g=0, b=128, name='blue')]
    >>>
    >>> Gradient(gradient + gradient[1::-1])
    >>> [RGB(r=128, g=0, b=0, name='red'), RGB(r=0, g=128, b=0, name='green'), RGB(r=0, g=0, b=128, name='blue'), RGB(r=0, g=128, b=0, name='green'), RGB(r=128, g=0, b=0, name='red')]

    """

    def __init__(self, colors: Sequence[ColorType] = ["pink1", "turquoise2"]):
        colors_ = [self._parse(color) for color in colors]
        super().__init__(colors_)

    def _parse(self, color: ColorType) -> RGB:
        if isinstance(color, RGB):
            return color
        elif isinstance(color, str):
            return RGB(name=color)
        elif isinstance(color, Tag):
            return RGB(name=color.name)
        return RGB(*color)

    def __setitem__(self, index: int, color: ColorType):  # type:ignore
        super().__setitem__(index, self._parse(color))

    def append(self, color: ColorType):
        super().append(self._parse(color))

    def insert(self, index: SupportsIndex, color: ColorType):
        super().insert(index, self._parse(color))

    def extend(self, colors: list[ColorType]):  # type:ignore
        super().extend([self._parse(color) for color in colors])

    def _get_blenders(self) -> list[_Blender]:
        return [_Blender(colors[0], colors[1]) for colors in itertools.pairwise(self)]

    def _batch_text(self, text: str, n: int) -> list[str]:
        """Split `text` into `n` chunks.

        All chunks will be the same size, except potentially the last chunk."""
        batch_size = int(len(text) / n)
        if batch_size == 0:
            return [ch for ch in text]
        batched_text = [
            text[i * batch_size : (i * batch_size) + batch_size] for i in range(n - 1)
        ]
        lastdex = n - 1
        batched_text.append(text[lastdex * batch_size :])
        return batched_text

    def apply(self, text: str) -> str:
        """Format `text` such that, when printed with `rich`,
        the displayed text changes colors according to this instance's color list."""
        blenders = self._get_blenders()
        batches = self._batch_text(text, len(blenders))
        return "".join(
            blender.apply(batch) for blender, batch in zip(blenders, batches)
        )



from .buds import ProgBar, Spinner, clear, print_in_place, ticker
from .colormap import ColorMap, Tag
from .gradient import RGB, Gradient
from .progress import (
    Progress,
    TimerColumn,
    get_bar_column,
    get_task_progress_column,
    track,
)

__version__ = "2.0.2"

__all__ = [
    "track",
    "Gradient",
    "ProgBar",
    "Spinner",
    "clear",
    "print_in_place",
    "ticker",
    "ColorMap",
    "Tag",
    "RGB",
    "Progress",
    "TimerColumn",
    "get_bar_column",
    "get_task_progress_column",
]



from datetime import timedelta
from typing import Any, Callable, Iterable, Optional, Self, Sequence

import rich.progress
from noiftimer import Timer
from rich.console import Console
from rich.progress import ProgressType
from rich.style import StyleType
from rich.text import Text

from .gradient import Gradient


def get_bar_column() -> rich.progress.BarColumn:
    return rich.progress.BarColumn(
        style="sea_green1",
        complete_style="deep_pink1",
        finished_style="cornflower_blue",
        pulse_style="deep_pink1",
    )


def get_task_progress_column(
    *args: Any, **kwargs: Any
) -> rich.progress.TaskProgressColumn:
    return rich.progress.TaskProgressColumn(
        "{task.percentage:>3.0f}%", style="light_coral", markup=False, *args, **kwargs
    )


class TimerColumn(rich.progress.TimeRemainingColumn):
    def __init__(self, elapsed_only: bool = False, *args: Any, **kwargs: Any):
        self.elapsed_only = elapsed_only
        super().__init__(*args, **kwargs)

    def get_time_remaining(self, task: rich.progress.Task) -> str:
        if self.elapsed_when_finished and task.finished:
            task_time = task.finished_time
        else:
            task_time = task.time_remaining
        if not task.total or not task_time:
            return ""
        time_remaining = Timer.format_time(task_time)
        if time_remaining == "<1s":
            time_remaining = "0s"
        return time_remaining

    def get_time_elapsed(self, task: rich.progress.Task) -> str:
        elapsed = task.finished_time if task.finished else task.elapsed
        if not elapsed:
            return ""
        delta = timedelta(seconds=max(0, int(elapsed)))
        time_elapsed = Timer.format_time(delta.total_seconds())
        if time_elapsed == "<1s":
            time_elapsed = "0s"
        return time_elapsed

    def render(self, task: rich.progress.Task) -> Text:
        timing = self.get_time_elapsed(task)
        if not self.elapsed_only and (time_remaining := self.get_time_remaining(task)):
            timing += f" <-> {time_remaining}"
            return Text().from_markup(Gradient().apply(timing))
        return Text().from_markup(f"[pink1]{timing}")


class Progress(rich.progress.Progress):
    """Renders an auto-updating progress bar(s).

    Args:
        console (Console, optional): Optional Console instance. Default will an internal Console instance writing to stdout.
        auto_refresh (bool, optional): Enable auto refresh. If disabled, you will need to call `refresh()`.
        refresh_per_second (Optional[float], optional): Number of times per second to refresh the progress information or None to use default (10). Defaults to None.
        speed_estimate_period: (float, optional): Period (in seconds) used to calculate the speed estimate. Defaults to 30.
        transient: (bool, optional): Clear the progress on exit. Defaults to False.
        redirect_stdout: (bool, optional): Enable redirection of stdout, so ``print`` may be used. Defaults to True.
        redirect_stderr: (bool, optional): Enable redirection of stderr. Defaults to True.
        get_time: (Callable, optional): A callable that gets the current time, or None to use Console.get_time. Defaults to None.
        disable (bool, optional): Disable progress display. Defaults to False
        expand (bool, optional): Expand tasks table to fit width. Defaults to False.

        description_last (bool, optional): When using the default columns, the description column will be after the bar instead of before.
    """

    def __enter__(self) -> Self:
        self.start()
        return self

    @classmethod
    def get_default_columns(cls) -> tuple[rich.progress.ProgressColumn, ...]:
        return (
            rich.progress.TextColumn("[pink1]{task.description}"),
            get_bar_column(),
            get_task_progress_column(),
            TimerColumn(),
            rich.progress.TextColumn("[pink1]{task.fields[suffix]}"),
        )

    def add_task(
        self,
        description: str = "",
        start: bool = True,
        total: float | None = 100,
        completed: int = 0,
        visible: bool = True,
        suffix: str = "",
        **fields: Any,
    ) -> rich.progress.TaskID:
        fields |= {"suffix": suffix}
        return super().add_task(description, start, total, completed, visible, **fields)


def track(
    sequence: Sequence[ProgressType] | Iterable[ProgressType],
    description: str = "Yeehaw...",
    total: Optional[float] = None,
    auto_refresh: bool = True,
    console: Optional[Console] = None,
    transient: bool = False,
    get_time: Optional[Callable[[], float]] = None,
    refresh_per_second: float = 10,
    style: StyleType = "sea_green1",
    complete_style: StyleType = "deep_pink1",
    finished_style: StyleType = "cornflower_blue",
    pulse_style: StyleType = "deep_pink1",
    update_period: float = 0.1,
    disable: bool = False,
    show_speed: bool = True,
) -> Iterable[ProgressType]:
    """Track progress by iterating over a sequence.

    Args:
        sequence (Iterable[ProgressType]): A sequence (must support "len") you wish to iterate over.
        description (str, optional): Description of task show next to progress bar. Defaults to "Working".
        total: (float, optional): Total number of steps. Default is len(sequence).
        auto_refresh (bool, optional): Automatic refresh, disable to force a refresh after each iteration. Default is True.
        transient: (bool, optional): Clear the progress on exit. Defaults to False.
        console (Console, optional): Console to write to. Default creates internal Console instance.
        refresh_per_second (float): Number of times per second to refresh the progress information. Defaults to 10.
        style (StyleType, optional): Style for the bar background. Defaults to "bar.back".
        complete_style (StyleType, optional): Style for the completed bar. Defaults to "bar.complete".
        finished_style (StyleType, optional): Style for a finished bar. Defaults to "bar.finished".
        pulse_style (StyleType, optional): Style for pulsing bars. Defaults to "bar.pulse".
        update_period (float, optional): Minimum time (in seconds) between calls to update(). Defaults to 0.1.
        disable (bool, optional): Disable display of progress.
        show_speed (bool, optional): Show speed if total isn't known. Defaults to True.
    Returns:
        Iterable[ProgressType]: An iterable of the values in the sequence.

    """

    columns = list(Progress.get_default_columns())
    progress = Progress(
        *columns,
        auto_refresh=auto_refresh,
        console=console,
        transient=transient,
        get_time=get_time,
        refresh_per_second=refresh_per_second or 10,
        disable=disable,
    )

    with progress:
        yield from progress.track(
            sequence, total=total, description=description, update_period=update_period
        )
